/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1995 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	rfc822.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	941008	extracted from nntp.c
	950312	encode/decode parts in a multipart message
	951029	extracted from mime.c of DeleGate
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
extern char *strstr();
extern char *strcasestr();
extern char *strdup();
extern char *MimeKit_Version;


char *findField(head,field,value)
	char *head,*field,**value;
{	char *nsp,*csp,*fsp;
	int flen;

	flen = strlen(field);
	for( csp = head; *csp; csp = nsp ){
		if( strncasecmp(csp,field,flen) == 0 ){
			fsp = csp;
			csp += flen;
			while( *csp == ' ' ) csp++;
			if( *csp == ':' ) csp++;
			while( *csp == ' ' ) csp++;
			if( value )
				*value = csp;
			return fsp;
		}
		if( nsp = strchr(csp,'\n') )
			nsp++;
		else	return 0;
	}
	return 0;
}

rmField(head,field)
	char *head,*field;
{	char *fp,*np;
	int nf;

	nf = 0;
	while( fp = findField(head,field,NULL) ){
		if( np = strchr(fp,'\n') )
			strcpy(fp,np+1);
		else	*fp = 0;
		nf++;
	}
	return nf;
}

char *findFieldValue(head,field)
	char *head,*field;
{	char *value;

	if( findField(head,field,&value) )
		return value;
	else	return NULL;
}

char *getFieldValue(head,field,value)
	char *head,*field,*value;
{	char ch,*vp,*bp;

	if( vp = findFieldValue(head,field) ){
		bp = value;
		while( ch = *vp++ ){
			if( ch == '\r' )
				continue;

			if( ch == '\n' ){
				if( *vp == ' ' || *vp == '\t' ){
					/*vp++;*/
					continue;
				}
				break;
			}
			if( ch == '\t' )
				ch = ' ';
			*bp++ = ch;
		}
		*bp = 0;
		return value;
	}
	value[0] = 0;
	return NULL;
}

char *fgetsHeaderField(hfp,name,value,size)
	FILE *hfp;
	char *name,*value;
{	char line[1024],*vp;
	int len;
	int found;
	int off;

	len = strlen(name);
	found = 0;
	value[0] = 0;

	off = ftell(hfp);
	for(;;){
		if( fgets(line,sizeof(line),hfp) == NULL )
			break;
		if( line[0] == '\r' || line[0] == '\n' )
			break;
		if( strncasecmp(line,name,len) == 0 ){
			vp = line + len;
			while( *vp == ' ' )
				vp++;
			if( *vp == ':' ){
				found++;
				vp++;
				linescan(vp,value,size);
			}
		}
	}
	fseek(hfp,off,0);

	if( found )
		return value;
	else	return NULL;
}

static strsubstr(n1,n2)
	char *n1,*n2;
{
	return strstr(n1,n2) == n1 || strstr(n2,n1) == n2;
}
int (*NNTP_nodematch)() = strsubstr;

findXref(afp,matchfunc,host,xref,size)
	FILE *afp;
	int (*matchfunc)();
	char *host,*xref;
{	char line[1024],node[256];
	char *dp,dc,*ep;
	int off,found;

	if( matchfunc == NULL )
		matchfunc = strsubstr;

	xref[0] = 0;
	off = ftell(afp);
	found = 0;

	while( fgets(line,sizeof(line),afp) != NULL ){
		if( line[0] == '\r' || line[0] == '\n' )
			break;
		if( line[0] == '.' )
		if( line[1] == '\r' || line[1] == '\n' )
			break;

		if( isspace(line[0]) )
			continue;

		if( (dp = strchr(line,':')) == 0 )
			continue;
		*dp++ = 0;
		if( strcasecmp(line,"Xref") != 0 )
			continue;

		while( *dp && isspace(*dp) )
			dp++;

		wordscan(dp,node);
		if( !(*matchfunc)(host,node) ) 
			continue;

		if( ep = strpbrk(dp,"\r\n") )
			*ep = 0; /* must treat folded line ... */

		strncpy(xref,dp,size-1);
		xref[size-1] = 0;
		found = 1;
		break;
	}
	fseek(afp,off,0);
	return found;
}

selectXref(host,xref1,xref2)
	char *host,*xref1,*xref2;
{	char *sp,*dp,*xp,name[512],host1[512],ch;
	int getit;
	int len;

	len = strlen(host);
	sp = xref1;
	dp = xref2;
	getit = 0;

	while( ch = *sp ){
		if( ch != ' ' && ch != '\t' ){
			getit = 0;
			if( strncasecmp(sp,"Xref",4) == 0 )
			if( xp = strchr(sp,':') ){
				wordscan(xp+1,host1);

				if( strncasecmp(host1,host,len) == 0 )
				if( host1[len] == 0 || host1[len] == '.' )
					getit = 1;
			}
		}
		while( ch = *sp ){
			sp++;
			if( getit )
				*dp++ = ch;
			if( ch == '\n' )
				break;
		}
	}
	*dp = 0;
}

static char *separator(field)
	char *field;
{
	if( strcasecmp(field,"To")==0 || strcasecmp(field,"Cc")==0 )
		return ", ";
	if( strcasecmp(field,"Received") == 0 )
		return "; ";
	return " ";
}

RFC821_skipheader(afp,out,field)
	FILE *afp,*out;
	char *field;
{	char line[1024],*lp,*dp,selected[1024];
	int flen,rcode,nputs;

	rcode = EOF;
	selected[0] = 0;
	nputs = 0;
	if( field != NULL )
		flen = strlen(field);
	else	flen = 0;

	while( fgets(line,sizeof(line),afp) != NULL ){
		if( line[0] == '\r' || line[0] == '\n' ){
			rcode = 0;
			break;
		}
		if( line[0] == '.' )
		if( line[1] == '\r' || line[1] == '\n' )
			break;

		if( out != NULL ){
		    if( field == NULL )
			fputs(line,out);
		    else{
			if( selected[0] == 0 ){
				if( line[flen] != ':' )
					continue;
				if( strncasecmp(line,field,flen) != 0 )
					continue;
				for( lp = line+flen+1; *lp; lp++ )
					if( !isspace(*lp) )
						break;
				if( nputs++ != 0 )
					fputs(separator(field),out);
				if( dp = strpbrk(line,"\r\n\f") )
					*dp = 0;
				fputs(lp,out);
			}else{
				if( isspace(line[0]) ){
					fputs(" ",out);
					fputs(line+1,out);
				}else	selected[0] = 0;
			}
		    }
		}
	}
	return rcode;
}
RFC821_skipbody(afp,out,line,size)
	FILE *afp,*out;
	char *line;
{	char linebuf[1024];
	int rcc;

	if( line == NULL ){
		line = linebuf;
		size = sizeof(linebuf);
	}
	*line = 0;
	rcc = 0;
	for(;;){
		if( fgets(line,size,afp) == NULL ){
			line[0] = 0;
			break;
		}
		if( line[0] == '.' )
		if( line[1] == '\r' || line[1] == '\n' )
			break;

		rcc += strlen(line);
		if( out != NULL )
			fputs(line,out);
	}
	return rcc;
}

char *readHeader(in)
	FILE *in;
{	char *lp,buff[0x10000];

	lp = buff;
	for(;;){
		if( fgets(lp,1024,in) == NULL )
			break;
		if( lp[0] == '\r' || lp[0] == '\n' )
			break;
		lp += strlen(lp);
	}
	return strdup(buff);
}

static charset_namelen(name)
	char *name;
{	char *np;
	int nc;

	for( np = name; nc = *np; np++ )
		if( nc != '-' && !isalnum(nc) )
			break;
	return np - name;
}

myMIMEversion(ver)
	char *ver;
{
	sprintf(ver,"1.0 (generated by MimeKit/%s)",MimeKit_Version);
}

char *strSeekEOH(head)
	char *head;
{	char *hp,ch;
	int top;

	top = 1;
	for( hp = head; ch = *hp; hp++ ){
		if( top ){
			if( ch == '\r' || ch == '\n' )
				break;
			if( ch == '.' && (hp[1] == '\r' || hp[1] == '\n') )
				break;
		}
		top = (ch == '\n');
	}
	return hp;
}
insert_ctype(head,ctype)
	char *head,*ctype;
{	char *ctp;
	char EOL[4],ver[128],tmp[0x10000];

	tmp[0] = 0;
	if( strstr(head,"\r\n") )
		strcpy(EOL,"\r\n");
	else	strcpy(EOL,"\n");

	ctp = strSeekEOH(head);
	strcpy(tmp,ctp);

	if( findFieldValue(head,"MIME-Version") == NULL ){
		myMIMEversion(ver);
		sprintf(ctp,"MIME-Version: %s%s",ver,EOL);
		ctp += strlen(ctp);
	}
	sprintf(ctp,"Content-Type: %s%s",ctype,EOL);
	strcat(ctp,tmp);
}

replace_charset(head,charset)
	char *head,*charset;
{	char *ctp,*csp,*cst;
	char tmp[0x10000];

	if( charset == NULL )
		return 0;

	if( (ctp = findFieldValue(head,"Content-Type")) == NULL ){
		char ctype[1024];
		sprintf(ctype,"text/plain; charset=%s",charset);
		insert_ctype(head,ctype);
		return 0;
	}

	if( csp = strcasestr(ctp,"charset=") ){
		csp += strlen("charset=");
		if(*csp == '"')
			csp++;
		if( strncasecmp(csp,charset,strlen(charset)) == 0 )
			return 0;
		cst = csp + charset_namelen(csp);
		strcpy(tmp,cst);
		strcpy(csp,charset);
		strcat(csp,tmp);
	}else{
		if( (cst = strpbrk(ctp,";\r\n")) == 0 )
			return 0;
		strcpy(tmp,cst);
		sprintf(cst,"; charset=%s",charset);
		strcat(cst,tmp);
	}
	return 1;
}

replaceFieldValue(head,field,value)
	char *head,*field,*value;
{	char *esp;
	char tmp[0x10000];
	char ch,*np;

	if( (esp = findFieldValue(head,field)) == NULL ){
		np = strSeekEOH(head);
		strcpy(tmp,np);
		sprintf(np,"%s: %s\r\n%s",field,value,tmp);
		return 0;
	}

	if( strncasecmp(esp,value,strlen(value)) == 0 )
		return 0;

	for( np = esp; ch = *np; np++ )
		if( ch == '\r' || ch == '\n' )
			break;
	strcpy(tmp,np);
	strcpy(esp,value);
	strcat(esp,tmp);
	return 1;
}

replaceContentType(head,type)
	char *head,*type;
{	char *ctp,*cst;
	char tmp[0x10000],*tp;

	ctp = findFieldValue(head,"Content-Type");
	if( ctp == NULL ){
		tp = strdup(head);
		sprintf(head,"Content-Type: %s\r\n%s",type,tp);
		free(tp);
		return 0;
	}
	if( (cst = strpbrk(ctp,";\r\n")) == 0 )
		return -1;
	strcpy(tmp,cst);
	strcpy(ctp,type);
	strcat(ctp,tmp);
	return 1;
}
